# Hash
import time
import hashlib
import hmac
from requests.auth import AuthBase
from urllib.parse import urlparse
from urllib import parse

# clients
import bitmex

# Utils
from dateutil import parser
import pytz
import pandas as pd
import math
import requests
import json
from dateutil import parser
from pandas.tseries.offsets import DateOffset
import pickle
import datetime as dt
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import warnings
warnings.simplefilter("ignore")

import os


from napoleontoolbox.file_saver import dropbox_file_saver

# Utils
def convert_unix(ts):
    return dt.datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')


def convert_string(s):
    dt_object = parser.parse(s)
    return dt_object.timestamp()


def assess_rsp(response):
    if response.status_code != 200:
        print('Bad gateway')
    elif isinstance(response.json(), list):
        pass
    # elif response.json()['Response'] != 'Success':
    # raise RuntimeError(response.json()['Message'])


def to_iso(ts):
    if type(ts) == str:
        ts = parser.parse(ts)
        return ts.replace(tzinfo=dt.timezone.utc).isoformat()
    else:
        return dt.datetime.utcfromtimestamp(ts).replace(tzinfo=dt.timezone.utc).isoformat()


def to_unix(s):
    if isinstance(s, str):
        dt_object = parser.parse(s)
    else:
        dt_object = s
    return int(dt_object.replace(tzinfo=dt.timezone.utc).timestamp())


def assess_rsp(response):
    if response.status_code != 200:
        raise RuntimeError('Bad gateway:', response.status_code)
    elif isinstance(response.json(), list) and len(response.json()) == 0:
        raise ValueError('No data')
    # elif response.json()['Response'] != 'Success':
    # raise RuntimeError(response.json()['Message'])


def extract_df(optype, precision, r):
    if optype == 'OHLC':
        try:
            tmp = r.json()['Data']['Data']
            return pd.DataFrame(tmp)
        except KeyError:
            print(r.json()['Message'])
    else:
        return pd.DataFrame(r.json()['Data'])


def convert_to_BitmexTS(s):
    s = parser.parse(s)
    s = s.strftime('%Y-%m-%d%%20%H%%3A%M')
    return s


# CryptoCompare
class CryptoCompare:
    def __init__(self, api_key=None, exchange=None):
        self.api_key = api_key
        self.e = exchange

    def get(self, optype, currency, startdate, enddate, precision):
        timedelta = to_unix(enddate) - to_unix(startdate)
        ts = to_unix(enddate)
        precision_dct = {'1h': 3600, 'hour': 3600, 'minute': 60}  # https://min-api.cryptocompare.com/data/v2/histohour
        endpoint_dct = {'OHLC': {'url': 'https://min-api.cryptocompare.com/data/v2/histo{}'.format(precision),
                                 'params': {'fsym': currency, 'tsym': 'USD', 'limit': '2000', 'aggreggate': '1',
                                            'toTs': ts}},
                        'OBL2': {'url': 'https://min-api.cryptocompare.com/data/ob/l2/snapshot',
                                 'params': {'api_key', self.api_key}},
                        'HVOL': {'url': 'https://min-api.cryptocompare.com/data/symbol/histohour',
                                 'params': {'fsym': currency, 'tsym': 'USD', 'limit': '500', 'toTs': ts}}}

        if optype == 'OHLC' and precision == 'minute':
            endpoint_dct[optype]['params']['api_key'] = '{' + self.api_key + '}'

        runs, rest = divmod(timedelta / precision_dct[precision], int(endpoint_dct[optype]['params']['limit']))
        runs, rest = int(runs), str(int(math.ceil(rest)))
        output = pd.DataFrame()
        for run in range(runs):
            r = requests.request("GET", endpoint_dct[optype]['url'], params=endpoint_dct[optype]['params'])
            assess_rsp(r)
            output = pd.concat([output, extract_df(optype, precision, r)])
            endpoint_dct[optype]['params'].update({'toTs': output.time.min()})

        if rest != '0':
            endpoint_dct[optype]['params'].update({'limit': rest})
            print(endpoint_dct[optype]['params'])
            r = requests.request("GET", endpoint_dct[optype]['url'], params=endpoint_dct[optype]['params'])
            assess_rsp(r)
            output = pd.concat([output, extract_df(optype, precision, r)])

        output['timestamp'] = output.time.apply(lambda x: to_iso(x))
        output = output.set_index('timestamp', drop=True).sort_index().drop_duplicates()
        return output


def generate_signature(secret, verb, url, expires, data):
    """Generate a request signature compatible with BitMEX."""
    # Parse the url so we can remove the base and extract just the path.
    parsedURL = urlparse(url)
    path = parsedURL.path
    if parsedURL.query:
        path = path + '?' + parsedURL.query

    if isinstance(data, (bytes, bytearray)):
        data = data.decode('utf8')

    # print("Computing HMAC: %s"  % verb + path + str(expires) + data)
    message = verb + path + str(expires) + data

    signature = hmac.new(bytes(secret, 'utf8'), bytes(message, 'utf8'), digestmod=hashlib.sha256).hexdigest()
    return signature


class APIKeyAuthWithExpires:  # (AuthBase):

    """Attaches API Key Authentication to the given Request object. This implementation uses `expires`."""

    def __init__(self, apiKey, apiSecret):
        self.apiKey = apiKey
        self.apiSecret = apiSecret

    def __call__(self, r):
        # modify and return the request
        expires = int(round(time.time()) + 5)  # 5s grace period in case of clock skew
        r.headers['api-expires'] = str(expires)
        r.headers['api-key'] = self.apiKey
        r.headers['api-signature'] = generate_signature(self.apiSecret, r.method, r.url, expires, r.data or '')

        return r


def gen_url_BitMEX(optype, currency='', startdate=convert_unix(time.time()), enddate=convert_unix(time.time()),
                   precision='', count=''):  # Ajouter un truc pour le format...
    startdate, enddate = convert_to_BitmexTS(startdate), convert_to_BitmexTS(enddate)
    url_base = 'https://www.bitmex.com/api/v1'
    op_cnvt_runs = {'position': {'endpoint': '/position?',
                                 'params': {'columns': 'account',
                                            'count': '50'}},
                    'stats': {'endpoint': '/stats',
                              'params': None},
                    'OBL2': {'endpoint': '/orderBook/L2?',
                             'params': {'symbol': currency,
                                        'depth': '50'}},
                    'OHLC': {'endpoint': '/trade/bucketed?',
                             'params': {'binSize': precision,
                                        'partial': 'false',
                                        'symbol': currency,
                                        'count': str(count),
                                        'reverse': 'true',
                                        'endTime': enddate}}}

    op_cnvt_rest = {'position': {'endpoint': '/position?',
                                 'params': {'columns': 'account',
                                            'count': '50'}},
                    'stats': {'endpoint': '/stats',
                              'params': None},
                    'OBL2': {'endpoint': '/orderBook/L2?',
                             'params': {'symbol': currency,
                                        'depth': '50'}},
                    'OHLC': {'endpoint': '/trade/bucketed?',
                             'params': {'binSize': precision,
                                        'partial': 'false',
                                        'symbol': currency,
                                        'count': '500',
                                        'reverse': 'true',
                                        'startTime': startdate,
                                        'endTime': enddate}}}

    if count == 500:
        op_cnvt = op_cnvt_runs
    else:
        op_cnvt = op_cnvt_rest

    if op_cnvt[optype]['params'] is not None:
        return url_base + op_cnvt[optype]['endpoint'] + '&'.join(
            ['='.join(list(item)) for item in op_cnvt[optype]['params'].items()])
    else:
        return url_base + op_cnvt[optype]['endpoint']


class BitMEX:

    def __init__(self, api_key, private_key, *args):
        self.api_key = api_key
        self.p_key = private_key
        self.KeyAuth = APIKeyAuthWithExpires(api_key, self.p_key)

    # Where optype = position, stats, funding, orderBook, trade, /user/executionHistory
    # Format we need for startdate, enddate = 2019-04-25T00:00:00.545Z

    def retrieve_and_assess(self, s, url):

        request = self.KeyAuth(requests.Request('GET', url))
        prepped = request.prepare()
        resp = s.send(prepped)
        assess_rsp(resp)
        temp = pd.DataFrame(resp.json())
        return temp

    def get(self, optype, currency='', startdate=convert_unix(time.time()), enddate=convert_unix(time.time()),
            precision='nearest'):

        import tqdm
        s = requests.Session()
        startdate_unix, enddate_unix = convert_string(startdate), convert_string(enddate)
        timedelta = abs(enddate_unix - startdate_unix)
        div_dct = {'daily': 3600 * 24, '1d': 3600 * 24, '1h': 3600, '1m': 60}
        if precision not in div_dct.keys():
            if precision == 'nearest' and optype == 'funding':
                div = 3600 * 8
            else:
                div = 60
        else:
            div = div_dct[precision]
        runs, rest = divmod(timedelta / div, 500)
        runs = int(runs)

        output = pd.DataFrame()
        if (optype != 'OBL2') and (optype != 'stats') and (runs != 0):
            for run in tqdm.tqdm(range(runs)):
                url = gen_url_BitMEX(optype, currency, startdate, enddate, precision, 500)
                output = pd.concat([output, self.retrieve_and_assess(s, url)], ignore_index=True)
                enddate = str(output['timestamp'].iloc[-1])
                time.sleep(3)

            if rest != 0:
                url = gen_url_BitMEX(optype, currency, startdate, enddate, precision)
                output = pd.concat([output, self.retrieve_and_assess(s, url)], ignore_index=True)

        else:
            url = gen_url_BitMEX(optype, currency, startdate, enddate, precision)
            output = self.retrieve_and_assess(s, url)

        output = output.drop_duplicates()
        return output


import numba
from numba import jit


def trades_to_arr(df):
    # Buy = 1
    # Sell = -1
    # XBT = 1
    # ETH = -1
    # Market = 1
    # Limit = 0
    # Funding = 0
    # Limit = 1
    from numba import typed
    df = df.replace('Buy', 1).replace('Sell', -1).replace('ETHUSD', -1).replace('XBTUSD', 1).replace('Limit',
                                                                                                     0).replace(
        'Market', 1).replace('Trade', 1).replace('Funding', 0)
    df = df[['symbol', 'side', 'lastQty', 'lastPx', 'orderQty', 'price', 'execType', 'ordType', 'leavesQty', 'cumQty',
             'avgPx', 'commission', 'execCost', 'execComm', 'homeNotional', 'foreignNotional', 'transactTime']].copy()
    df.transactTime = df.transactTime.apply(lambda x: x.timestamp() * 1000)
    df = df.replace(np.nan, 0)
    df = df.replace({'(.*?)': 0}, regex=True)
    d = typed.Dict()
    for i, k in enumerate(df.columns):
        d[k] = i
    return df.astype(float).values, d


def trades_to_arr(df):
    # Buy = 1
    # Sell = -1
    # XBT = 1
    # ETH = -1
    # Market = 1
    # Limit = 0
    # Funding = 0
    # Limit = 1
    from numba import typed
    df = df.replace('Buy', 1).replace('Sell', -1).replace('ETHUSD', -1).replace('XBTUSD', 1).replace('Limit',
                                                                                                     0).replace(
        'Market', 1).replace('Trade', 1).replace('Funding', 0)
    df = df[['symbol', 'side', 'lastQty', 'lastPx', 'orderQty', 'price', 'execType', 'ordType', 'leavesQty', 'cumQty',
             'avgPx', 'commission', 'execCost', 'execComm', 'homeNotional', 'foreignNotional', 'transactTime']].copy()
    df.transactTime = df.transactTime.apply(lambda x: x.timestamp() * 1000)
    df = df.replace(np.nan, 0)
    df = df.replace({'(.*?)': 0}, regex=True)
    d = typed.Dict()
    for i, k in enumerate(df.columns):
        d[k] = i
    return df.astype(float).values, d


@jit(nopython=True)
def find_lastPx(arr, d):
    lastPxBTC = []
    lastPxETH = []
    for i in range(arr.shape[0]):
        if arr[i, d['symbol']] == 1:
            lastPxBTC.append(arr[i, d['lastPx']])
            t = list(arr[:i, d['symbol']][::-1])
            if -1 not in t:
                lastPxETH.append(np.nan)
            else:
                r = t.index(-1)
                lastPxETH.append(arr[i - r - 1, d['lastPx']])
        else:
            lastPxETH.append(arr[i, d['lastPx']])
            t = list(arr[:i, d['symbol']][::-1])
            if 1 not in t:
                lastPxBTC.append(np.nan)
            else:
                r = t.index(1)
                lastPxBTC.append(arr[i - r - 1, d['lastPx']])

    return np.array(lastPxBTC), np.array(lastPxETH)

    # if len(l[0]) == 0:
    # return np.nan
    # else:
    # return l


@jit(nopython=True)
def get_qtraded(arr, d):
    qtraded_ETH = [187465]  # Faut aller voir dans la feuille de trades sinon c nop
    qtraded_BTC = [620484]
    for i in range(1, arr.shape[0]):
        if (arr[i, d['symbol']] == 1) and (arr[i, d['execType']] == 1):
            qtraded_ETH.append(0)
            qtraded_BTC.append(arr[i, d['lastQty']])
        elif (arr[i, d['symbol']] == -1) and (arr[i, d['execType']] == 1):
            qtraded_BTC.append(0)
            qtraded_ETH.append(arr[i, d['lastQty']])
        else:
            qtraded_BTC.append(0)
            qtraded_ETH.append(0)
    return np.array(qtraded_BTC), np.array(qtraded_ETH)


@jit(nopython=True)
def get_sizes(arr, d):
    size_ETH = []
    size_BTC = []
    for i in range(arr.shape[0]):
        if (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == -1):
            size_ETH.append(arr[i, d['lastPx']] * arr[i, d['lastQty']])
            size_BTC.append(0)
        elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == 1):
            size_BTC.append(arr[i, d['lastPx']] * arr[i, d['lastQty']])
            size_ETH.append(0)
        else:
            size_ETH.append(0)
            size_BTC.append(0)
    return np.array(size_BTC), np.array(size_ETH)


@jit(nopython=True)
def get_senses(arr, d):
    sens_ETH = []
    sens_BTC = []
    for i in range(arr.shape[0]):
        if (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == -1) and (arr[i, d['side']] == 1):
            sens_ETH.append(1.)
            sens_BTC.append(0.)
        elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == -1) and (arr[i, d['side']] == -1):
            sens_ETH.append(-1.)
            sens_BTC.append(0)
        elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == 1) and (arr[i, d['side']] == 1):
            sens_ETH.append(0)
            sens_BTC.append(1.)
        elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == 1) and (arr[i, d['side']] == -1):
            sens_ETH.append(0.)
            sens_BTC.append(-1.)
        else:
            sens_ETH.append(0.)
            sens_BTC.append(0.)
    sens_BTC = np.array(sens_BTC)
    sens_ETH = np.array(sens_ETH)
    return sens_BTC, sens_ETH


@jit(nopython=True)
def get_markets(arr, d):
    markets_BTC = []
    markets_ETH = []
    for i in range(arr.shape[0]):
        if (arr[i, d['symbol']] == 1) and (arr[i, d['execType']] == 1) and (arr[i, d['ordType']] == 1):
            markets_BTC.append(arr[i, d['lastQty']])
            markets_ETH.append(0)
        elif (arr[i, d['symbol']] == -1) and (arr[i, d['execType']] == 1) and (arr[i, d['ordType']] == 1):
            markets_ETH.append(arr[i, d['lastQty']])
            markets_BTC.append(0)
        else:
            markets_ETH.append(0)
            markets_BTC.append(0)
    markets_BTC = np.array(markets_BTC)  # .astype(float)
    markets_ETH = np.array(markets_ETH)  # .astype(float)
    return markets_BTC, markets_ETH


@jit(nopython=True)
def _merge_update(arr, d, arrays: list, names: list):
    if len(arrays) != len(names):
        return None
    else:
        last_idx = arr.shape[1] + 1
        for a in arrays:
            t = a.reshape(-1, 1)
            arr = np.hstack((arr, t))
        for _ in names:
            d[_] = int(last_idx)
            last_idx += 1
    return arr, d


@jit(nopython=True)
def numba_compute(arr, d):
    arrays = []
    a, b = find_lastPx(arr, d)
    arrays.extend([a, b])
    a, b = get_qtraded(arr, d)
    arrays.extend([a, b])
    a, b = get_sizes(arr, d)
    arrays.extend([a, b])
    a, b = get_senses(arr, d)
    arrays.extend([a, b])
    a, b = get_markets(arr, d)
    arrays.extend([a, b])
    names = ['lastPxBTC', 'lastPxETH', 'traded_xbt', 'traded_eth', 'size_xbt', 'size_eth', 'sens_trade_xbt',
             'sens_trade_eth', 'markets_BTC', 'markets_ETH']
    arr, d = _merge_update(arr, d, arrays, names)
    #arrays = get_fees(df, d)
    arrays = get_fees(arr, d)

    names = ['funding_fees_xbt_btc', 'funding_fees_eth_btc', 'funding_fees_xbt_usd', 'funding_fees_eth_usd',
             'trading_fees_xbt_btc', 'trading_fees_xbt_usd', 'trading_fees_eth_btc', 'trading_fees_eth_usd']
    arr, d = _merge_update(arr, d, arrays, names)
    return arr, d


@jit(nopython=True)
def get_fees(arr, d):
    if 'lastPxBTC' not in d:
        return None
    else:
        funding_fees_xbt_btc = []
        funding_fees_eth_btc = []
        funding_fees_xbt_usd = []
        funding_fees_eth_usd = []
        trading_fees_xbt_btc = []
        trading_fees_xbt_usd = []
        trading_fees_eth_btc = []
        trading_fees_eth_usd = []
        fold = [funding_fees_xbt_btc, funding_fees_eth_btc, funding_fees_xbt_usd, funding_fees_eth_usd,
                trading_fees_xbt_btc, trading_fees_xbt_usd, trading_fees_eth_btc, trading_fees_eth_usd]
        for i in range(arr.shape[0]):
            if (arr[i, d['execType']] == 0) and (arr[i, d['symbol']] == 1):
                ff = arr[i, d['execComm']] / 100000000
                funding_fees_xbt_btc.append(ff)
                funding_fees_xbt_usd.append(ff * arr[i, d['lastPxBTC']])
                max_l = funding_fees_xbt_btc
            elif (arr[i, d['execType']] == 0) and (arr[i, d['symbol']] == -1):
                ff = arr[i, d['execComm']] / 100000000
                funding_fees_eth_btc.append(ff)
                funding_fees_eth_usd.append(ff * arr[i, d['lastPxBTC']])
                max_l = funding_fees_eth_btc
            elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == 1):
                tf = arr[i, d['execComm']] / 100000000
                trading_fees_xbt_btc.append(tf)
                trading_fees_xbt_usd.append(tf * arr[i, d['lastPxBTC']])
                max_l = trading_fees_xbt_btc
            elif (arr[i, d['execType']] == 1) and (arr[i, d['symbol']] == -1):
                tf = arr[i, d['execComm']] / 100000000
                trading_fees_eth_btc.append(tf)
                trading_fees_eth_usd.append(tf * arr[i, d['lastPxBTC']])
                max_l = trading_fees_eth_btc
            for _ in fold:
                if len(_) < len(max_l):
                    _.append(0)

        fold_a = []
        for _ in fold:
            fold_a.append(np.array(_))
        return fold_a

def fetch_all_necessary_data(public_key = None, private_key= None, start_time= None, end_time= None, save_to_dropbox = False, dropbox_token = None, local_root_directory = './'):

    dbx = dropbox_file_saver.NaPoleonDropboxConnector(drop_token=dropbox_token, dropbox_backup=True)

    start_time_stub = start_time.strftime('%d_%b_%Y')
    end_time_stub = end_time.strftime('%d_%b_%Y')
    filename_stub = f'{public_key}_{start_time_stub}_{end_time_stub}_Bitmex'
    #filename_stub = f'{inv_year_dictionary[year]}_{inv_month_dictionary[month]}_{inv_market_dictionary[market]}'


    report_pkl_file_name = f'{filename_stub}_report.pkl'
    trade_pkl_file_name = f'{filename_stub}_trades.pkl'
    ohlc_eth_bm_pkl_file_name = f'{filename_stub}_ohlc_eth_bm.pkl'
    ohlc_btc_bm_pkl_file_name = f'{filename_stub}_ohlc_btc_bm.pkl'
    walletHist_pkl_file_name = f'{filename_stub}_walletHist.pkl'
    ohlc_btc_cc_pkl_file_name = f'{filename_stub}_ohlc_btc_cc.pkl'
    ohlc_eth_cc_pkl_file_name = f'{filename_stub}_ohlc_eth_cc.pkl'

    client = bitmex.bitmex(test=False, config=None, api_key=public_key, api_secret=private_key)
    if isinstance(start_time,dt.datetime) and isinstance(end_time,dt.datetime):
        start_time = start_time.strftime('%Y-%m-%dT%H:%M')
        end_time = end_time.strftime('%Y-%m-%dT%H:%M')


    ce, te = parser.parse(start_time).replace(tzinfo=pytz.UTC), parser.parse(end_time).replace(tzinfo=pytz.UTC)

    ce = ce - dt.timedelta(hours=8)
    trades = pd.DataFrame()
    print('Accessing Trade History...')
    while ce < te:
        print(f'requesting trades between {ce} and {te}')
        tmp = pd.DataFrame(client.Execution.Execution_getTradeHistory(startTime=ce, count=500).result()[0]).sort_values(
            by='timestamp', ascending=True)
        trades = trades.append(tmp)
        time.sleep(4)
        ce = tmp.timestamp.iloc[-1]
    trades = trades[trades.timestamp <= te].copy()
    trades.execComm = - trades.execComm  # logique comptable
    trades = trades.sort_values(by='timestamp', ascending=True).drop_duplicates()
    print('Done.')

    # Get wallet history for a period of time
    ce, te = parser.parse(start_time).replace(tzinfo=pytz.UTC), parser.parse(end_time).replace(tzinfo=pytz.UTC)
    ce = ce - dt.timedelta(hours=8)
    walletHist = pd.DataFrame()
    print('Accessing wallet history...')
    while ce < te:
        print(f'requesting wallet history between {ce} and {te}')
        tmp = pd.DataFrame(client.User.User_getWalletHistory(currency='XBt').result()[0]).sort_values(by='timestamp',
                                                                                                      ascending=True)
        walletHist = walletHist.append(tmp)
        time.sleep(3.5)
        ce = tmp.timestamp.iloc[-1]
    walletHist = walletHist[walletHist.timestamp <= te].copy()
    print('Wallet History retrieved.')

    # Récupérer les OHLC hourly sur Cryptocompare depuis startdate
    cc = CryptoCompare('')
    ohlc_btc_cc = cc.get('OHLC', 'BTC', start_time, end_time, precision='hour')
    ohlc_eth_cc = cc.get('OHLC', 'ETH', start_time, end_time, precision='hour')
    print("OHLC Cryptocompare retrieved.")

    # Récupérer les OHLC hourly sur BM depuis startdate

    bm = BitMEX(public_key, private_key)
    ohlc_btc_bm = bm.get('OHLC', currency='XBT', startdate=start_time, enddate=end_time, precision='1h')
    ohlc_eth_bm = bm.get('OHLC', currency='ETH', startdate=start_time, enddate=end_time, precision='1h')
    print('OHLC BM retrieved.')

    if save_to_dropbox :
        full_path = local_root_directory + trade_pkl_file_name
        trades.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=trade_pkl_file_name, overwrite=True)
        os.remove(full_path)

        full_path = local_root_directory + ohlc_eth_bm_pkl_file_name
        ohlc_eth_bm.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=ohlc_eth_bm_pkl_file_name, overwrite=True)
        os.remove(full_path)

        full_path = local_root_directory + ohlc_btc_bm_pkl_file_name
        ohlc_btc_bm.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=ohlc_btc_bm_pkl_file_name, overwrite=True)
        os.remove(full_path)

        full_path = local_root_directory + walletHist_pkl_file_name
        walletHist.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=walletHist_pkl_file_name, overwrite=True)
        os.remove(full_path)

        full_path = './' + ohlc_btc_cc_pkl_file_name
        ohlc_btc_cc.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=ohlc_btc_cc_pkl_file_name, overwrite=True)
        os.remove(full_path)

        full_path = local_root_directory + ohlc_eth_cc_pkl_file_name
        ohlc_eth_cc.to_pickle(full_path)
        print(f'uploading to dropbox {full_path}')
        dbx.upload(fullname=full_path, folder='', subfolder='', name=ohlc_eth_cc_pkl_file_name, overwrite=True)
        os.remove(full_path)

    # ohlc_btc_bm
    x = ohlc_btc_bm
    x = x.set_index('timestamp')
    x.index = x.index.map(parser.parse)
    x.index = x.index.tz_localize(None)
    x.columns = [y + '_btc_bm' for y in x.columns]
    ohlc_btc_bm = x

    # ohlc_eth_bm
    x = ohlc_eth_bm
    x = x.set_index('timestamp')
    x.index = x.index.map(parser.parse)
    x.index = x.index.tz_localize(None)
    x.columns = [y + '_eth_bm' for y in x.columns]
    ohlc_eth_bm = x

    # ohlc_btc_cc
    x = ohlc_btc_cc
    x.index = x.index.map(parser.parse)
    x.index = x.index.tz_localize(None)
    x.columns = [y + '_btc_cc' for y in x.columns]
    ohlc_btc_cc = x

    # ohlc_eth_cc
    x = ohlc_eth_cc
    x.index = x.index.map(parser.parse)
    x.index = x.index.tz_localize(None)
    x.columns = [y + '_eth_cc' for y in x.columns]
    ohlc_eth_cc = x

    ohlc_eth_cc = ohlc_eth_cc.sort_index(ascending=True).drop_duplicates()
    ohlc_btc_cc = ohlc_btc_cc.sort_index(ascending=True).drop_duplicates()
    ohlc_btc_bm = ohlc_btc_bm.sort_index(ascending=True).drop_duplicates()
    ohlc_eth_bm = ohlc_eth_bm.sort_index(ascending=True).drop_duplicates()

    ##############################
    ##############################
    ##########
    ########## computing, grouping
    ##########
    ##############################
    ##############################

    start = time.time()
    df, d = trades_to_arr(trades)
    df, d = numba_compute(df, d)
    print('Done for', trades.shape[0], 'trades in ', time.time() - start, 'seconds.')

    # Now: Groupby
    df = pd.DataFrame(df, columns=[i for i in d.keys()])
    df['funding_fees_xbt_usd'] = df.funding_fees_xbt_btc * df.lastPxBTC
    df['funding_fees_eth_usd'] = df.funding_fees_eth_btc * df.lastPxBTC
    df['trading_fees_eth_usd'] = df.trading_fees_eth_btc * df.lastPxBTC
    df['trading_fees_xbt_usd'] = df.trading_fees_xbt_btc * df.lastPxBTC
    df.transactTime = df.transactTime.apply(lambda x: dt.datetime.utcfromtimestamp(x / 1000))
    df['dt_floor'] = df.transactTime.dt.floor('H')
    df[
        'total_fees_usd'] = df.trading_fees_xbt_usd + df.trading_fees_eth_usd + df.funding_fees_xbt_usd + df.funding_fees_eth_usd

    print('Grouping by...')
    td = df[['dt_floor', 'traded_eth', 'traded_xbt', 'size_eth', 'size_xbt', 'sens_trade_eth', 'sens_trade_xbt',
             'funding_fees_xbt_usd', 'funding_fees_eth_usd',
             'trading_fees_xbt_btc', 'trading_fees_xbt_usd', 'trading_fees_eth_btc', 'trading_fees_eth_usd',
             'total_fees_usd']].groupby('dt_floor').sum()

    td.index = td.index.tz_localize(None)
    td['avg_eth_price'] = td.size_eth / abs(td.traded_eth)
    td['avg_btc_price'] = td.size_xbt / abs(td.traded_xbt)

    # Join & compute prices for cryptocompare
    td = td.join(ohlc_eth_cc['open_eth_cc']).join(ohlc_btc_cc['open_btc_cc'])
    td['slippage_eth_price_cc'] = (td.avg_eth_price / td.open_eth_cc - 1) * np.sign(td.sens_trade_eth)
    td['slippage_eth_cash_cc'] = td.slippage_eth_price_cc * abs(
        td.traded_eth)  # / 1000000 * td.open_btc_cc * td.open_eth_cc
    td['slippage_btc_price_cc'] = (td.avg_btc_price / td.open_btc_cc - 1) * np.sign(td.sens_trade_xbt)
    td['slippage_btc_cash_cc'] = td.slippage_btc_price_cc * abs(td.traded_xbt)
    # slippage en 100e de millionnième de eth = ratio * cours open * qty * 100
    # slippage en 100e de millionnième de btc = ratio * 1Mo/ cours * 100 * qty

    # Join & compute slippage for bitmex
    td = td.join(ohlc_eth_bm['open_eth_bm']).join(ohlc_btc_bm['open_btc_bm'])
    td['slippage_eth_price_bm'] = (td.avg_eth_price / td.open_eth_bm - 1) * np.sign(td.sens_trade_eth)
    td['slippage_eth_cash_bm'] = td.slippage_eth_price_bm * abs(
        td.traded_eth)  # / 1000000 * td.open_btc_bm * td.open_eth_bm
    td['slippage_btc_price_bm'] = (td.avg_btc_price / td.open_btc_bm - 1) * np.sign(td.sens_trade_xbt)
    td['slippage_btc_cash_bm'] = td.slippage_btc_price_bm * abs(td.traded_xbt)
    print('All done.')

    walletHist = walletHist.set_index('timestamp')
    walletHist.index = walletHist.index.map(lambda x: x.floor('60s'))
    walletHist = walletHist[~walletHist.index.duplicated(keep='first')].sort_index(ascending=True)
    walletHist.index = walletHist.index.tz_localize(None)
    walletHist['walletBalance_xbt'] = walletHist.walletBalance / 100000000
    walletHist = walletHist.join(ohlc_btc_cc['open_btc_cc'])
    walletHist['walletBalance_usd'] = walletHist.walletBalance_xbt * walletHist.open_btc_cc

    # Join to td
    td_h = td.join(walletHist.walletBalance_usd.dropna(), rsuffix='_fromWH')
    td_h.walletBalance_usd = td_h.walletBalance_usd.fillna(method='ffill')

    td_h['traded_eth_usd_a'] = td_h.traded_eth * td_h.avg_eth_price * 0.000001 * td_h.open_btc_bm
    td_h['traded_xbt_usd_a'] = td_h.traded_xbt / td_h.avg_btc_price
    # Usd value = xbt spot * eht usd price * bitcoin multiplier ( 0.000001)

    x = td_h[['traded_eth', 'avg_eth_price', 'traded_eth_usd_a', 'sens_trade_eth', 'trading_fees_eth_usd',
              'funding_fees_eth_usd', 'open_eth_cc', 'open_eth_bm',
              'traded_xbt', 'avg_btc_price', 'traded_xbt_usd_a', 'sens_trade_xbt', 'trading_fees_xbt_usd',
              'funding_fees_xbt_usd', 'open_btc_cc', 'open_btc_bm', 'walletBalance_usd',
              'trading_fees_eth_usd', 'funding_fees_eth_usd', 'slippage_eth_price_cc', 'slippage_eth_cash_cc',
              'trading_fees_xbt_usd', 'funding_fees_xbt_usd', 'slippage_btc_price_cc', 'slippage_btc_cash_cc',
              'trading_fees_eth_usd', 'funding_fees_eth_usd', 'slippage_eth_price_bm', 'slippage_eth_cash_bm',
              'trading_fees_xbt_usd', 'funding_fees_xbt_usd', 'slippage_btc_price_bm', 'slippage_btc_cash_bm']]

    x = x.rename_axis('timestamp')
    x.columns = ['quantity_traded_eth', 'avg_traded_price_eth', 'amount_traded_usd_eth', 'n_trades_eth',
                 'trading_fees_usd_eth', 'funding_fees_usd_eth',
                 'open_eth_cc', 'open_eth_exchange',
                 'quantity_traded_btc', 'avg_traded_price_btc', 'amount_traded_usd_btc', 'n_trades_btc',
                 'trading_fees_usd_btc', 'funding_fees_usd_btc',
                 'open_btc_cc', 'open_btc_exchange',
                 'wallet_amount_usd',
                 'trading_fees_usd_eth', 'funding_fees_usd_eth', 'slippage_eth_price_cc', 'slippage_eth_cash_cc',
                 'trading_fees_usd_btc', 'funding_fees_usd_btc', 'slippage_btc_price_cc', 'slippage_btc_cash_cc',
                 'trading_fees_usd_eth', 'funding_fees_usd_eth', 'slippage_eth_price_bm', 'slippage_eth_cash_bm',
                 'trading_fees_usd_btc', 'funding_fees_usd_btc', 'slippage_btc_price_bm', 'slippage_btc_cash_bm']

    x = x.fillna(0).round(decimals=5)

    full_path = local_root_directory + report_pkl_file_name
    x.to_pickle(full_path)
    print(f'uploading to dropbox {full_path}')
    dbx.upload(fullname=full_path, folder='', subfolder='', name=report_pkl_file_name, overwrite=True)
    os.remove(full_path)

    return x




from __future__ import print_function

def init_napoleon_colab():
    import os
    stream = os.popen('pip install dropbox')
    output = stream.read()
    print(output)

    stream = os.popen('pip install plotly==4.7.1')
    output = stream.read()
    print(output)

    stream = os.popen('pip install dash == 1.13.3')
    output = stream.read()
    print(output)

    from napoleontoolbox.signal import signal_utility
    from napoleontoolbox.analyzer import minutely_market, hourly_market, market
    from datetime import datetime
    import pandas as pd

    from ipywidgets import interact, interactive, fixed, interact_manual
    import ipywidgets as widgets
    from napoleontoolbox.utility import metrics

    from pathlib import Path

    import matplotlib.pyplot as plt


    import plotly.offline as py
    py.init_notebook_mode(connected=True)
    import plotly.graph_objs as go
    import plotly
    from plotly.offline import iplot



    def configure_plotly_browser_state():
      import IPython
      display(IPython.core.display.HTML('''
            <script src="/static/components/requirejs/require.js"></script>
            <script>
              requirejs.config({
                paths: {
                  base: '/static/base',
                  plotly: 'https://cdn.plot.ly/plotly-latest.min.js?noext',
                },
              });
            </script>
            '''))
    import IPython

    configure_plotly_browser_state()
    IPython.get_ipython().events.register('pre_run_cell', configure_plotly_browser_state)





import nltk
from nltk.tokenize import sent_tokenize, word_tokenize
from nltk.tag import pos_tag
from nltk.stem.wordnet import WordNetLemmatizer
from nltk.corpus import stopwords
import itertools


def sentence_tokenizer(text):
    tokens = [word_tokenize(t) for t in sent_tokenize(text)]
    return tokens

def lemmatize_sentence(tokens):
    lemmatizer = WordNetLemmatizer()
    lemmatized_sentence = []
    for word, tag in pos_tag(tokens):
        if tag.startswith('NN'):
            pos = 'n'
        elif tag.startswith('VB'):
            pos = 'v'
        else:
            pos = 'a'
        lemmatized_sentence.append(lemmatizer.lemmatize(word, pos))
    return lemmatized_sentence

def pre_cleaning_sentences(text=''):


    #Lowercasing
    lower_text = text.lower()
    # Tokenize
    tokens = sentence_tokenizer(lower_text)

    # Lemmatizing
    lemma_text = [lemmatize_sentence(token) for token in tokens]
    lemma_text = list(itertools.chain.from_iterable(lemma_text))

    # Removing stop words
    tokens_without_sw = [word for word in lemma_text if not word in stopwords.words()]

    # Removing special characters
    alnum_token_list = [''.join(e for e in S if e.isalnum()) for S in tokens_without_sw]

    processed_tokens = list(filter(lambda a: a != '', alnum_token_list))
    return processed_tokens


#!/usr/bin/env python
# coding: utf-8


def generate_pairs_trading_simple_corrle(look_backs, keep_same_position_periods_list, correlation_threshold_list, suffixes_list):
    parameters = []
    for look_back in look_backs:
        for keep_same_position_periods in keep_same_position_periods_list:
            for correlation_threshold in correlation_threshold_list:
                for suffixes in suffixes_list:
                    parameters.append({
                        'lookback_window': look_back,
                        'keep_same_position_periods': keep_same_position_periods,
                        'correlation_threshold': correlation_threshold,
                        'suffixes' : suffixes})
    return parameters

def generate_pair_trading_shifted_correlation(look_backs, correlation_ranges, keep_same_position_periods_list, variants_list, suffixes_list):
    parameters = []
    for look_back in look_backs:
        for correlation_range in correlation_ranges:
            for suffixes in suffixes_list:
                for keep_same_position_periods in keep_same_position_periods_list:
                    for variant in variants_list:
                        if look_back > correlation_range:
                            parameters.append({
                                'lookback_window': look_back,
                                'correlation_range' : correlation_range,
                                'keep_same_position_periods': keep_same_position_periods,
                                'variant': variant,
                                'suffixes': suffixes})
    return parameters

def pair_trading_shifted_correlation_with_cointegration(look_backs, correlation_ranges, keep_same_position_periods_list, risk_list, alpha_list, variants_list, suffixes_list):
    parameters = []
    for look_back in look_backs:
        for correlation_range in correlation_ranges:
            for suffixes in suffixes_list:
                for keep_same_position_periods in keep_same_position_periods_list:
                    for risk in risk_list:
                        for alpha in alpha_list:
                            for variant in variants_list:
                                if look_back > correlation_range:
                                    parameters.append({
                                        'lookback_window': look_back,
                                        'correlation_range' : correlation_range,
                                        'keep_same_position_periods': keep_same_position_periods,
                                        'suffixes': suffixes,
                                        'risk': risk,
                                        'alpha': alpha,
                                        'variant': variant})
    return parameters



def generate_pairs_trading_log_simple_baseline_parameters(look_backs, keep_same_position_periods_list, percent_deviations, risk_list, alpha_list, suffixes_list):
    parameters = []
    for look_back in look_backs:
        for keep_same_position_periods in keep_same_position_periods_list:
            for percent_deviation in percent_deviations:
                for risk in risk_list:
                    for alpha in alpha_list:
                        for suffixes in suffixes_list:
                            parameters.append({
                                'lookback_window': look_back,
                                'keep_same_position_periods': keep_same_position_periods,
                                'percent_deviation': percent_deviation,
                                'risk': risk,
                                'alpha': alpha,
                                'suffixes' : suffixes})
    return parameters

def generate_pairs_trading_log_zscore_baseline_parameters(look_back_windows, keep_same_position_periods_list, risk_list, alpha_list, skipping_point_list, threshold_type_list, suffixes_list, number_of_zscore_dev_list):
    parameters = []
    for look_back_window in look_back_windows:
        for keep_same_position_periods in keep_same_position_periods_list:
            for risk in risk_list:
                for alpha in alpha_list:
                    for skipping_point in skipping_point_list:
                        for threshold_type in threshold_type_list:
                            for suffixes in suffixes_list:
                                for number_of_zscore_dev in number_of_zscore_dev_list:
                                    if skipping_point < look_back_window/2:
                                        parameters.append({
                                            'lookback_window': look_back_window,
                                            'keep_same_position_periods': keep_same_position_periods,
                                            'risk': risk,
                                            'alpha': alpha,
                                            'skipping_point': skipping_point,
                                            'threshold_type': threshold_type,
                                            'suffixes': suffixes,
                                            'number_of_zscore_dev': number_of_zscore_dev})
    return parameters

def generate_pairs_trading_log_ecm_baseline_a(look_back_windows, keep_same_position_periods_list, risk_list, alpha_list, mean_lag_list, ecm_shift_list, suffixes_list):
    parameters = []
    for look_back_window in look_back_windows:
        for keep_same_position_periods in keep_same_position_periods_list:
            for risk in risk_list:
                for alpha in alpha_list:
                    for mean_lag in mean_lag_list:
                        for ecm_shift in ecm_shift_list:
                            for suffixes in suffixes_list:
                                if mean_lag < 0.7 * look_back_window:
                                    parameters.append({
                                        'lookback_window': look_back_window,
                                        'keep_same_position_periods': keep_same_position_periods,
                                        'risk': risk,
                                        'alpha': alpha,
                                        'mean_lag': mean_lag,
                                        'ecm_shift': ecm_shift,
                                        'suffixes': suffixes})
    return parameters





__author__ = "hugo.inzirillo"

import requests


class Get(object):
    def __init__(self, flag):
        self.flag = flag
        self.connector = None
    def __call__(self, original_func):
        def wrappee(connector, **kwargs):
            self.connector = connector
            original_func(self.flag)
        return wrappee


    def __fill_request(self):
        self.connector


class Connector(object):
    def __init__(self):
        self.__hello = "Hugo"

    @property
    def hello(self):
        return self.__hello

    @Get("https://www.bitmex.com/api/v1/orderBook/L2?symbol=xbt&depth=100000")
    def get_prices(self, **kwargs):
        return requests.get(**kwargs)


if __name__ == '__main__':
    a = Connector().get_prices(url="coucou")



__author__ = "hugo.inzirillo"

from dataclasses import dataclass
from typing import List, Union
from datetime import time
from pandas import Timestamp, Timedelta
from itertools import product

@dataclass
class Product:
    productCode: str = None
    label: str = None
    currency: str = None
    productType: str = None
    stratType: str = None
    basketProfile: str = None
    basketReturnType: str = None
    rebalancingPeriod: str = None
    inceptionDate: str = None
    initialValue: float = None
    underlyings: List[str] = None
    market: str = None
    creationTs: str = None
    groups: List[str] = None
    maturity: str = None
    isFrontFuture: bool = None
    futurePeriodicity: str = None
    underlyingFuture: str = None
    rollNDaysBeforeMaturity: int = None
    targetVol: float = None
    guaranteedRate: float = None
    maxLeverage: float = None
    numberOfDecimalsForCalculation: int = None
    numberOfDecimalsForDisplay: str = None
    stratCron: str = None
    expired: bool = None

    def from_dict(self, dict_object: dict):
        if isinstance(dict_object, dict):
            self.__dict__ = dict_object
            return self

    def __repr__(self):
        return self.productCode

    def __hash__(self):
        return hash(self.productCode) if self.productCode is not None else 0

    def __eq__(self, other):
        return self.productCode == other.productCode if isinstance(other,
                                                                   self.__class__) else False


@dataclass
class ProductView:
    productCode: str = None
    label: str = None
    stratType: str = None
    underlyings: List[str] = None

    def from_dict(self, dict_object: dict):
        if isinstance(dict_object, dict):
            self.__dict__ = {key: value for key, value in dict_object.items()
                             if key in set(dict_object.keys()).intersection(set(self.__dict__.keys()))}
            return self

    def __repr__(self):
        return self.productCode

    def __hash__(self):
        return hash(self.productCode) if self.productCode is not None else 0


@dataclass
class EodQuoteProductCodeBuilder:
    underlying: str
    date: [Union[str, Timestamp, Timedelta]]
    product_code: str = None

    def __post_init__(self):
        if self.product_code is None and isinstance(self.date, str):
            self.product_code = self.underlying + "-" + str(Timestamp(self.date).hour) + "H"
        if self.product_code is None and isinstance(self.date, Timestamp):
            self.product_code = self.underlying + "-" + str(self.date.hour) + "H"
        if self.product_code is None and isinstance(self.date, time):
            self.product_code = self.underlying + "-" + str(self.date.hour) + "H"


    @staticmethod
    def product_code_generator(udl_code: List[str]):
        product_code_combination = product(udl_code, range(0, 24))
        return [EodQuoteProductCodeBuilder(product_code, time(hour=hour)).product_code for product_code, hour in
                product_code_combination]


    def __hash__(self):
        return hash(self.product_code)

    def __eq__(self, other):
        return self.product_code == other if isinstance(other, self.__class__) else False


if __name__ == '__main__':
    test = ProductView().from_dict(Product(productCode="Hello", label="Test", stratType=",elf,ezof,zeof,zef").__dict__)
    # ProductView(productCode=prod.productCode, label=prod.label, stratType=prod.stratType, underlyings=prod.underlyings)



__author__ = "hugo.inzirillo"

from dataclasses import dataclass
from typing import List

from arch.univariate.volatility import GARCH
from pandas import Timestamp

from napoleontoolbox.backend.quote_service import NapoleonQuoteService
from napoleontoolbox.models.filter import EodQuoteFilter
from napoleontoolbox.models.product import EodQuoteProductCodeBuilder

filter = EodQuoteFilter(productCodes=EodQuoteProductCodeBuilder.product_code_generator(["ETH-USD", "BTC-USD"]),
                        minDate="2020-08-21",
                        maxDate="2020-08-24")
quotes = NapoleonQuoteService().filter(filter)


@dataclass(frozen=True)
class GarchEstimationParams(object):
    start_date: Timestamp = None
    end_date: Timestamp = None
    underlyings: List[str] = None
    p: int = 1
    o: int = 0
    q: int = 1


@dataclass(frozen=True)
class GarchEstimationResult(object):
    pass


class GarchEstimation(object):
    def __init__(self, params: GarchEstimationParams):
        self.__model = GARCH
        self.__quotes = list()
        self.__parameters = params

    def __post_init__(self):
        self.__load_data()

    @property
    def model(self):
        return self.__model

    @property
    def parameters(self):
        return self.__parameters

    def __load_data(self):
        if self.__parameters.start_date and self.__parameters.end_date and self.__parameters.underlyings:
            filter = EodQuoteFilter(
                productCodes=EodQuoteProductCodeBuilder.product_code_generator(self.__parameters.underlyings),
                minDate=self.__parameters.start_date.strftime("%Y-%m-%d"),
                maxDate=self.__parameters.start_date.strftime("%Y-%m-%d"))
            quotes = NapoleonQuoteService().filter(filter)
            if quotes is not None:
                self.__quotes = quotes
            else:
                raise Exception("Quote is none type : credentials")
        else:
            raise Exception("Missing parameters : check start_date ; end_date ; underlyings")


if __name__ == '__main__':
    params = GarchEstimationParams(
        start_date=Timestamp("2020-08-21"),
        end_date=Timestamp("2020-08-24"),
        underlyings=["ETH-USD", "BTC-USD"]
    )
    test = GarchEstimation(params)



#!/usr/bin/env python3
# coding: utf-8

import sqlite3

from multiprocessing import Pool
from napoleontoolbox.file_saver import dropbox_file_saver
from napoleontoolbox.parallel_run import launcher_utility

from napoleontoolbox.signal import signal_utility
import json

import math

import numpy as np

class AugmentoForecasterParalleLauncher():
    def __init__(self,starting_date = None, running_date = None, drop_token='', dropbox_backup=True, local_root_directory='../data',user='napoleon', underlying=None, frequence=None, db_path_suffix = '_forecaster.sqlite', advanced_features = [False],ss=[10], ns=[21], s_evals = [5], calibration_steps=[1000],  signal_idiosyncratic_paramaters=[None], table_max_size = 2000):
        self.starting_date = starting_date
        self.running_date = running_date
        self.args = []
        self.key_args=[]
        self.counter = 1
        self.seed = 0
        self.mapping_keys={}
        for adv_feat in advanced_features:
            for s in ss:
                for n in ns:
                    for s_eval in s_evals:
                        for calibration_step in calibration_steps:
                            for signal_type, params  in signal_idiosyncratic_paramaters.items():
                                for param in params:
                                    self.args.append((self,self.seed,adv_feat,n,s,s_eval,calibration_step,signal_type,json.dumps(param, sort_keys=True)))
                                    self.key_args.append((self.seed,adv_feat,n,s,s_eval,calibration_step,signal_type,json.dumps(param, sort_keys=True)))
                                    self.counter = self.counter + 1
        self.key_args.sort()
        self.counter = 1
        for me_arg in self.key_args:
            for me_arg in self.key_args:
                continuous_saving_key, continuous_saving_key_scale_one, continuous_saving_key_scale_two, two_states_saving_key, two_states_saving_key_lo, three_states_saving_key, three_states_saving_key_lo, _ = self.create_saving_key(
                    *me_arg)
                self.mapping_keys[continuous_saving_key] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[continuous_saving_key_scale_one] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[continuous_saving_key_scale_two] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[two_states_saving_key] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[two_states_saving_key_lo] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[three_states_saving_key] = self.counter
                self.counter = self.counter + 1
                self.mapping_keys[three_states_saving_key_lo] = self.counter
                self.counter = self.counter + 1

        self.run_number = self.counter - 1
        self.local_root_directory = local_root_directory
        self.user = user

        self.db_path_suffix = db_path_suffix
        self.filename =  launcher_utility.get_database_name(user, frequence, underlying, 'default', starting_date, running_date, db_path_suffix)
        self.db_path = self.local_root_directory + self.filename
        self.runs = []
        self.totalRow = 0

        self.table_max_size=table_max_size
        self.nb_tables = math.ceil(self.run_number/self.table_max_size)

        self.instantiateTables()
        self.instantiateSignalTables()
        self.empty_runs_to_investigate = []
        self.dbx = dropbox_file_saver.NaPoleonDropboxConnector(drop_token=drop_token,dropbox_backup=dropbox_backup)

    def create_saving_key(self, seed, adv_feat, n, s, s_eval, calibration_step, signal_type, idios_string):
        idios = json.loads(idios_string)
        common_params = {
            'seed': seed,
            'adv_feat':adv_feat,
            'n': n,
            's': s,
            's_eval': s_eval,
            'calibration_step': calibration_step,
            'signal_type': signal_type,
            'discretization': 0,
            'type': 'continuous'
        }
        common_params.update(idios)
        continuous_saving_key = json.dumps(common_params, sort_keys=True)
        continuous_saving_key = signal_utility.convert_to_sql_column_format(continuous_saving_key)

        common_params.update({
            'discretization': 1,
            'type': 'long short'
        })
        two_states_saving_key = json.dumps(common_params, sort_keys=True)
        two_states_saving_key = signal_utility.convert_to_sql_column_format(two_states_saving_key)

        common_params.update({
            'discretization': 2,
            'type': 'long only'
        })
        two_states_saving_key_lo = json.dumps(common_params, sort_keys=True)
        two_states_saving_key_lo = signal_utility.convert_to_sql_column_format(two_states_saving_key_lo)

        common_params.update({
            'discretization': 3,
            'type': 'long short'
        })
        three_states_saving_key = json.dumps(common_params, sort_keys=True)
        three_states_saving_key = signal_utility.convert_to_sql_column_format(three_states_saving_key)

        common_params.update({
            'discretization': 4,
            'type': 'long only'
        })
        three_states_saving_key_lo = json.dumps(common_params, sort_keys=True)
        three_states_saving_key_lo = signal_utility.convert_to_sql_column_format(three_states_saving_key_lo)

        common_params.update({
            'discretization': -1,
            'type': 'long short'
        })
        continuous_saving_key_scale_one = json.dumps(common_params, sort_keys=True)
        continuous_saving_key_scale_one = signal_utility.convert_to_sql_column_format(continuous_saving_key_scale_one)

        common_params.update({
            'discretization': -2,
            'type': 'long short'
        })

        continuous_saving_key_scale_two = json.dumps(common_params, sort_keys=True)
        continuous_saving_key_scale_two = signal_utility.convert_to_sql_column_format(continuous_saving_key_scale_two)

        return continuous_saving_key, continuous_saving_key_scale_one, continuous_saving_key_scale_two, two_states_saving_key, three_states_saving_key_lo, three_states_saving_key, three_states_saving_key_lo, idios

    def launchParallelPool(self, toRun, use_num_cpu):
        with Pool(processes=use_num_cpu) as pool:
            results = pool.starmap(toRun, self.args)
            print('parallel computation done')

    def launchSequential(self, toRun):
        for meArg in self.args:
             toRun(*meArg)

    def instantiateSignalTables(self):
        for table_counter in range(self.nb_tables):
            self.instantiateSignalTable(table_counter)

    def instantiateSignalTable(self, table_counter):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = '''CREATE TABLE parallel_signal_'''+str(table_counter)+'''(
                                        effective_date date PRIMARY KEY);'''
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
            print('we do not recreate tables but resume the run with partially filled tables')
            print('if you want a new run, drop the table')
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")

    def instantiateTables(self):
        for table_counter in range(self.nb_tables):
            self.instantiateTable(table_counter)

    def instantiateTable(self, table_counter):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = '''CREATE TABLE parallel_run_'''+str(table_counter)+'''(
                                        effective_date date PRIMARY KEY);'''
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")


    def insertSignals(self, table_counter, run, values):
        sqliteConnection = None
        try:
            success = True
            sqliteConnection = sqlite3.connect(self.db_path)
            parallel_signal = 'parallel_signal_' + str(table_counter)
            cursor = sqliteConnection.cursor()
            for i, v in values.iteritems():
                sqlite_insert_query = """INSERT INTO '"""+parallel_signal+"""' ('effective_date', '""" + run + """') VALUES ('""" + str(
                    i) + """','""" + str(v) + """');"""
                cursor.execute(sqlite_insert_query)
            sqliteConnection.commit()
            cursor.close()
        except Exception as error:
            print("Error while working with SQLite", error)
            print('investigate')
            success = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")
        return success

    def insertResults(self, table_counter, run, values):
        sqliteConnection = None
        try:
            success = True
            sqliteConnection = sqlite3.connect(self.db_path)
            parallel_run = 'parallel_run_' + str(table_counter)
            cursor = sqliteConnection.cursor()
            for i, v in values.iteritems():
                sqlite_insert_query = """INSERT INTO '"""+parallel_run+"""' ('effective_date', '""" + run + """') VALUES ('""" + str(
                    i) + """','""" + str(v) + """');"""
                cursor.execute(sqlite_insert_query)
            sqliteConnection.commit()
            cursor.close()
        except Exception as error:
            print("Error while working with SQLite", error)
            print('investigate')

            success = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")
        return success

    def updateSignals(self, table_counter, run, values):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            cursor = sqliteConnection.cursor()
            parallel_signal = 'parallel_signal_' + str(table_counter)
            for i, v in values.iteritems():
                sqlite_update_query = """UPDATE '"""+parallel_signal+"""' set '""" + run + """' = '""" + str(
                    v) + """' where effective_date = '""" + str(i) + """'"""
                cursor.execute(sqlite_update_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
            print('investigate')
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")

    def updateResults(self, table_counter, run, values):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            parallel_run = 'parallel_run_' + str(table_counter)
            cursor = sqliteConnection.cursor()
            for i, v in values.iteritems():
                sqlite_update_query = """UPDATE '"""+parallel_run+"""' set '""" + run + """' = '""" + str(
                    v) + """' where effective_date = '""" + str(i) + """'"""
                cursor.execute(sqlite_update_query)
            sqliteConnection.commit()
            cursor.close()
        except sqlite3.Error as error:
            print("Error while working with SQLite", error)
            print('investigate')
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("The Sqlite connection is closed")

    def addResults(self, table_counter, run, values):
        success = self.insertResults(table_counter,run,values)
        if not success :
            self.updateResults(table_counter,run, values)

    def addSignals(self, table_counter, run, values):
        success = self.insertSignals(table_counter,run,values)
        if not success :
            self.updateSignals(table_counter,run, values)

    def addSignal(self, table_number, run):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = 'alter table parallel_signal_'+str(table_number)+' add column ' + run + ' real'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            self.runs.append(run)
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column

    def addRun(self, table_number, run):
        sqliteConnection = None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_create_table_query = 'alter table parallel_run_'+str(table_number)+' add column ' + run + ' real'
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_create_table_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            self.runs.append(run)
        except sqlite3.Error as error:
            print("Error while creating a sqlite table", error)
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column

    def getRunAssociatedTable(self, run):
        run_counter = self.mapping_keys[run]
        table_number = run_counter % self.nb_tables
        return table_number

    def saveAll(self,run, strat_results_df):
        table_number = self.getRunAssociatedTable(run)
        self.saveResults(table_number, run, strat_results_df['reconstituted_perf'])
        self.saveSignals(table_number, run, strat_results_df['signal'])

    def saveResults(self, table_number, run, strat_results_df):
        self.addRun(table_number,run)
        self.addResults(table_number, run, strat_results_df)

    def saveSignals(self, table_number, run, strat_results_df):
        self.addSignal(table_number, run)
        self.addSignals(table_number, run, strat_results_df)


    def checkLocalRunExistence(self, table_number, run):
        sqliteConnection = None
        runExistence=None
        try:
            sqliteConnection = sqlite3.connect(self.db_path)
            sqlite_select_run_query = 'SELECT effective_date,'+run + ' FROM parallel_run_'+str(table_number)
            cursor = sqliteConnection.cursor()
            print("Successfully Connected to SQLite")
            cursor.execute(sqlite_select_run_query)
            sqliteConnection.commit()
            print("SQLite table created")
            cursor.close()
            runExistence = True
        except sqlite3.Error as error:
            print("The run is not present", error)
            runExistence = False
        finally:
            if (sqliteConnection):
                sqliteConnection.close()
                print("sqlite connection is closed")  # adding the column
        return runExistence

    def exhaustiveCheckRunExistence(self,run):
        for table_number in range(self.nb_tables):
            if self.checkLocalRunExistence(table_number,run):
                return True, table_number
        return False, np.nan

    def checkRunExistence(self,run):
        table_number = self.getRunAssociatedTable(run)
        return self.checkLocalRunExistence(table_number,run), table_number

    def upload_results_to_dropbox(self):
        print('uploading to dropbox the results '+self.db_path)
        self.dbx.uploadFileToDropbox(filename=self.filename,fullpath = self.db_path)



__author__ = "hugo.inzirillo"

from itertools import groupby
from typing import List, Set, Dict, Callable

from pandas import DataFrame, Timestamp

from napoleontoolbox.models.filter import SignalComputationFilter
from napoleontoolbox.models.portfolio import Portfolio, PortfolioPk, PortfolioView
from napoleontoolbox.models.portfolio_decomposition import PortfolioDecomposition
from napoleontoolbox.models.product import Product, ProductView, EodQuoteProductCodeBuilder
from napoleontoolbox.models.quantity import Quantity
from napoleontoolbox.models.quote import EodQuotePk, EodQuote
from napoleontoolbox.models.request_portfolio_analysis import RequestPortfolioAnalysis
from napoleontoolbox.models.signal import SignalComputation, SignalComputationView
from napoleontoolbox.models.strats import StratTypeView


class SignalComputationFilterFactory(object):

    def to_signal_filer(self, product_codes: List[str], min_ts=None, max_ts=None, last_only: bool = True):
        return SignalComputationFilter(productCodes=product_codes, minTs=min_ts, maxTs=max_ts, lastOnly=last_only)

    def to_dict(self, signal_filter: SignalComputationFilter):
        return signal_filter.__dict__


class ProductFactory(object):

    def __from_set_to_dict(self, product_set: Set[Product]) -> Dict[str, Product]:
        return {item.productCode: item.__dict__ for item in product_set}

    def to_df(self, product_set: Set[Product]):
        return DataFrame.from_dict(data=self.__from_set_to_dict(product_set), orient="index")

    def groupby(self, iterable: Set[Product], key: Callable[[Product], str]):
        return {strat_type: set(_grouper) for strat_type, _grouper in
                groupby(sorted(iterable, key=key, reverse=True), key)}

    def to_view(self, product: Product):
        product_view = ProductView()
        product_view.stratType = StratTypeView.__getitem__(product.stratType).value
        product_view.label = product.label
        product_view.underlyings = product.underlyings
        product_view.productCode = product.productCode
        return product_view

    def to_product(self, product_view: ProductView):
        product = Product()
        product.productCode = None
        product.label = product_view.label
        product.stratType = product_view.stratType
        product.underlyings = product_view.underlyings
        return product


class SignalComputationFactory(object):
    def groupby(self, iterable: List[SignalComputation]
                , key: Callable[[SignalComputation], str]) -> Dict[str, List[SignalComputation]]:
        return {strat_type: set(_grouper) for strat_type, _grouper in
                groupby(sorted(iterable, key=key, reverse=True), key)}

    def to_view(self, signal: SignalComputation) -> SignalComputationView:
        return SignalComputationView(signal.pk.productCode, signal.pk.underlyingCode, signal.pk.ts, signal.value)


class QuantityFactory(object):
    def to_view(self, quantity: Quantity) -> SignalComputationView:
        return SignalComputationView(quantity.pk.productCode, quantity.pk.underlyingCode, quantity.pk.ts,
                                     quantity.value)


class RequestPortfolioAnalysisFactory(object):
    def to_portfolio(self, request: RequestPortfolioAnalysis) -> Portfolio:
        pk = PortfolioPk(request.amount, request.leverage, request.composition, request.start_date, request.end_date)
        portfolio = Portfolio(pk=pk, products=request.products)
        return portfolio


class PortfolioFactory(object):
    TS = "ts"
    UDL_CODE = "underlyingCode"
    VALUE = "value"

    def to_view(self, portfolio: Portfolio):
        return PortfolioView().from_dict(portfolio.__dict__)

    def to_portfolio_decomposition(self, porfolio: Portfolio, quotes: Dict[EodQuotePk, EodQuote]) -> List[
        PortfolioDecomposition]:
        res = list()
        agg_weights_by_ts = self.__get_aggregated_weight_values(porfolio.positions)
        agg_quantities_by_ts = self.__get_aggregated_quantity_values(porfolio.quantities)
        ts_udl = set(agg_quantities_by_ts.keys())
        for ts, udl in ts_udl:
            res.append(
                PortfolioDecomposition(ts, udl, agg_weights_by_ts.get((ts, udl)), agg_quantities_by_ts.get((ts, udl)),
                                       quotes.get(
                                           EodQuotePk(EodQuoteProductCodeBuilder(udl, self.to_date(ts)).product_code,
                                                      self.to_date(ts))).close
                                       )
            )

        return res

    def __get_aggregated_quantity_values(self, quantity_list: List[Quantity]) -> Dict:
        quantity_list = [QuantityFactory().to_view(quantity) for quantity in quantity_list]
        df = DataFrame(quantity_list)
        return \
            df.groupby([PortfolioFactory.TS, PortfolioFactory.UDL_CODE])[[PortfolioFactory.VALUE]].sum().to_dict()[
                PortfolioFactory.VALUE]

    def __get_aggregated_weight_values(self, signal_list: List[SignalComputation]) -> Dict:
        weight_list = [SignalComputationFactory().to_view(signal) for signal in signal_list]
        df = DataFrame(weight_list)
        return \
            df.groupby([PortfolioFactory.TS, PortfolioFactory.UDL_CODE])[[PortfolioFactory.VALUE]].sum().to_dict()[
                PortfolioFactory.VALUE]

    def to_date(self, ts: str) -> str:
        return Timestamp(ts).strftime("%Y-%m-%d")



__author__ = "hugo.inzirillo"

from dataclasses import dataclass

from pandas import Timestamp, Timedelta

from models.strats import StratType


@dataclass
class SignalComputationPk:
    productCode: str
    underlyingCode: str
    ts: str

    def __hash__(self):
        return hash((self.productCode, self.underlyingCode, self.ts))

    def __eq__(self, other):
        return self.productCode == other.productCode and self.underlyingCode == other.underlyingCode and \
               self.ts == other.ts if isinstance(other, self.__class__) else False

    def __ne__(self, other):
        return self.productCode != other.productCode and self.underlyingCode != other.underlyingCode and \
               self.ts != other.ts if isinstance(other, self.__class__) else True

    def __le__(self, other):
        return self.ts <= other.ts

    def __ge__(self, other):
        return self.ts >= other.ts

    def __lt__(self, other):
        return self.ts < other.ts

    def __gt__(self, other):
        return self.ts > other.ts


@dataclass
class SignalComputation:
    """
    Position Service
    Mapping of Java Objects
    """
    pk: SignalComputationPk = None
    value: float = None

    def from_dict(self, dict_object: dict, strat_type: StratType = None):
        if strat_type:
            if strat_type.value == StratType.AT_CLOSE_STRAT.value:
                dict_object["ts"] = (Timestamp(dict_object["ts"]) + Timedelta(hours=1)).strftime("%Y-%m-%dT%H:00:00")
            elif strat_type == StratType.CRON_STRAT:
                if strat_type.value == StratType.AT_CLOSE_STRAT.value:
                    dict_object["ts"] = Timestamp(dict_object["ts"]).strftime("%Y-%m-%dT%H:00:00")
        else:
            dict_object["ts"] = Timestamp(dict_object["ts"]).strftime("%Y-%m-%dT%H:00:00")

        self.pk = SignalComputationPk(dict_object["productCode"], dict_object["underlyingCode"], dict_object["ts"])
        self.value = dict_object["value"]
        return self

    def __hash__(self):
        return hash((self.pk))

    def __eq__(self, other):
        return self.pk == other.pk if isinstance(other, self.__class__) else False


@dataclass
class SignalComputationView:
    productCode: str = None
    underlyingCode: str = None
    ts: str = None
    value: float = None



from napoleontoolbox.rebalancing import rolling

from sklearn.metrics import mean_squared_error
from sklearn.metrics import accuracy_score
from sklearn.metrics import confusion_matrix

from napoleontoolbox.signal import signal_utility


import pandas as pd
import numpy as np



def rolling_mixing(forecasting_model, X, y, n=252, s=63, s_eval= None, calibration_step=None, display = False, **kwargs):
    assert X.shape[0] == y.shape[0]
    idx = X.index
    forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_three_states_forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_three_states_forecasting_series_lo = pd.Series(index=idx, name='prediction')
    discrete_two_states_forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_two_states_forecasting_series_lo = pd.Series(index=idx, name='prediction')

    if n is None and s_eval is None:
        roll = rolling._ExpandingRollingMechanism(idx, s=s)
    if n is None and s_eval is not None:
        roll = rolling._ExpandingEvalRollingMechanism(idx, s=s, s_eval = s_eval)
    if n is not None and s_eval is None :
        roll = rolling._RollingMechanism(idx, n=n, s=s)
    if n is not None and s_eval is not None:
        roll = rolling._EvalRollingMechanism(idx, n=n, s=s, s_eval = s_eval)
    features_names = list(X.columns)
    features_importances = []
    iteration_counter = 0
    for slice_n, slice_s, slice_s_eval in roll():
        # Select X
        X_train = X.loc[slice_n].copy()
        y_train = y.loc[slice_n].copy()

        X_test = X.loc[slice_s].copy()
        y_test = y.loc[slice_s].copy()
        if calibration_step >0:
            if (iteration_counter>0 and iteration_counter%calibration_step == 0):
                # print('launching calibration '+str(iteration_counter))
                forecasting_model.calibrate(X_train, y_train)
                # print('endiing calibration ' + str(iteration_counter))

        if slice_s_eval is not None:
            X_eval = X.loc[slice_s_eval].copy()
            y_eval = y.loc[slice_s_eval].copy()
            forecasting_model.fit(X_train, y_train, X_eval, y_eval)
        else :
            forecasting_model.fit(X_train, y_train, X_train, y_train)

        y_pred = forecasting_model.predict(X_test)
        y_train_pred = forecasting_model.predict(X_train)

        #### three states discretization
        lower_quantile_threshold, upper_quantile_threshold, lower_threshold, upper_threshold, sharpe_strat, sharpe_under = signal_utility.find_best_three_states_discretization(y_train_pred,y_train)

        y_pred_three_space_discrete = signal_utility.apply_discretization_three_states(y_pred, upper_threshold,lower_threshold)
        y_test_three_space_discrete = signal_utility.apply_discretization_three_states(y_test, upper_threshold,lower_threshold)

        #### three states discretization long only
        lower_quantile_threshold_lo, upper_quantile_threshold_lo, lower_threshold_lo, upper_threshold_lo, sharpe_strat_lo, sharpe_under_lo = signal_utility.find_best_three_states_discretization_lo(y_train_pred, y_train)


        y_pred_three_space_discrete_lo = signal_utility.apply_discretization_three_states_lo(y_pred, upper_threshold_lo,lower_threshold_lo)
        y_test_three_space_discrete_lo =signal_utility. discretize_three_states_lo(y_test, upper_threshold_lo,lower_threshold_lo)


        #### two states discretization


        splitting_quantile_threshold, splitting_threshold, sharpe_strat, sharpe_under = signal_utility.find_best_two_states_discretization(y_train_pred,y_train)
        y_pred_two_space_discrete = signal_utility.apply_discretization_two_states(y_pred, splitting_threshold)
        y_test_two_space_discrete = signal_utility.apply_discretization_two_states(y_test, splitting_threshold)

        #### two states discretization long only
        splitting_quantile_threshold_lo, splitting_threshold_lo, sharpe_strat_lo, sharpe_under_lo = signal_utility.find_best_two_states_discretization(y_train_pred,y_train)
        y_pred_two_space_discrete_lo = signal_utility.apply_discretization_two_states_lo(y_pred, splitting_threshold_lo)
        y_test_two_space_discrete_lo = signal_utility.apply_discretization_two_states_lo(y_test, splitting_threshold_lo)

        run_importances=forecasting_model.get_features_importance(features_names)
        features_importances.append(run_importances)
        # no leverage for the moment
        forecasting_series.loc[slice_s] = y_pred.ravel()
        discrete_three_states_forecasting_series.loc[slice_s] = y_pred_three_space_discrete.ravel()
        discrete_two_states_forecasting_series.loc[slice_s] = y_pred_two_space_discrete.ravel()
        discrete_three_states_forecasting_series_lo.loc[slice_s] = y_pred_three_space_discrete_lo.ravel()
        discrete_two_states_forecasting_series_lo.loc[slice_s] = y_pred_two_space_discrete_lo.ravel()

        rmse = mean_squared_error(y_test, y_pred)
        matrix_three = confusion_matrix(y_test_three_space_discrete, y_pred_three_space_discrete)
        accuracy_three = accuracy_score(y_test_three_space_discrete, y_pred_three_space_discrete)
        # matrix_three_lo = confusion_matrix(y_test_three_space_discrete_lo, y_pred_three_space_discrete_lo)
        # accuracy_three_lo = accuracy_score(y_test_three_space_discrete_lo, y_pred_three_space_discrete_lo)
        matrix_two = confusion_matrix(y_test_two_space_discrete, y_pred_two_space_discrete)
        accuracy_two = accuracy_score(y_test_two_space_discrete, y_pred_two_space_discrete)
        matrix_two_lo = confusion_matrix(y_test_two_space_discrete_lo, y_pred_two_space_discrete_lo)
        accuracy_two_lo = accuracy_score(y_test_two_space_discrete_lo, y_pred_two_space_discrete_lo)

        if display:
            print('training slice ' + str(slice_n))
            print('testing slice ' + str(slice_s))
            print('rmse for slice : '+str(rmse))
            print('accuracy three states')
            print(accuracy_three)
            print('confusion matrix three states')
            print(matrix_three  )
            print('accuracy two states')
            print(accuracy_two)
            print('confusion matrix two states')
            print(matrix_two)
        iteration_counter = iteration_counter+1

    features_importances = pd.DataFrame(features_importances)
    return forecasting_series, features_importances, discrete_two_states_forecasting_series, discrete_two_states_forecasting_series_lo, discrete_three_states_forecasting_series, discrete_three_states_forecasting_series_lo


def rolling_forecasting(forecasting_model, X, y, n=252, s=63, s_eval= None, calibration_step=None, display = False, min_signal = -1., max_signal = 1., **kwargs):
    assert X.shape[0] == y.shape[0]
    idx = X.index
    forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_three_states_forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_three_states_forecasting_series_lo = pd.Series(index=idx, name='prediction')
    discrete_two_states_forecasting_series = pd.Series(index=idx, name='prediction')
    discrete_two_states_forecasting_series_lo = pd.Series(index=idx, name='prediction')

    if n is None and s_eval is None:
        roll = rolling._ExpandingRollingMechanism(idx, s=s)
    if n is None and s_eval is not None:
        roll = rolling._ExpandingEvalRollingMechanism(idx, s=s, s_eval = s_eval)
    if n is not None and s_eval is None :
        roll = rolling._RollingMechanism(idx, n=n, s=s)
    if n is not None and s_eval is not None:
        roll = rolling._EvalRollingMechanism(idx, n=n, s=s, s_eval = s_eval)
    features_names = list(X.columns)
    features_importances = []
    iteration_counter = 0
    for slice_n, slice_s, slice_s_eval in roll():
        # Select X
        X_train = X.loc[slice_n].copy()
        y_train = y.loc[slice_n].copy()

        X_test = X.loc[slice_s].copy()
        y_test = y.loc[slice_s].copy()
        if calibration_step >0:
            if (iteration_counter>0 and iteration_counter%calibration_step == 0):
                # print('launching calibration '+str(iteration_counter))
                forecasting_model.calibrate(X_train, y_train)
                # print('endiing calibration ' + str(iteration_counter))

        if slice_s_eval is not None:
            X_eval = X.loc[slice_s_eval].copy()
            y_eval = y.loc[slice_s_eval].copy()
            forecasting_model.fit(X_train, y_train, X_eval, y_eval)
        else :
            forecasting_model.fit(X_train, y_train, X_train, y_train)

        y_pred = forecasting_model.predict(X_test)
        y_train_pred = forecasting_model.predict(X_train)

        #### three states discretization
        lower_quantile_threshold, upper_quantile_threshold, lower_threshold, upper_threshold, sharpe_strat, sharpe_under = signal_utility.find_best_three_states_discretization(
            y_train_pred, y_train)

        y_pred_three_space_discrete = signal_utility.apply_discretization_three_states(y_pred, upper_threshold,
                                                                                       lower_threshold)
        y_test_three_space_discrete = signal_utility.apply_discretization_three_states(y_test, upper_threshold,
                                                                                       lower_threshold)

        #### three states discretization long only
        lower_quantile_threshold_lo, upper_quantile_threshold_lo, lower_threshold_lo, upper_threshold_lo, sharpe_strat_lo, sharpe_under_lo = signal_utility.find_best_three_states_discretization_lo(
            y_train_pred, y_train)

        y_pred_three_space_discrete_lo = signal_utility.apply_discretization_three_states_lo(y_pred, upper_threshold_lo,
                                                                                             lower_threshold_lo)
        y_test_three_space_discrete_lo = signal_utility.discretize_three_states_lo(y_test, upper_threshold_lo,
                                                                                   lower_threshold_lo)

        #### two states discretization

        splitting_quantile_threshold, splitting_threshold, sharpe_strat, sharpe_under = signal_utility.find_best_two_states_discretization(
            y_train_pred, y_train)
        y_pred_two_space_discrete = signal_utility.apply_discretization_two_states(y_pred, splitting_threshold)
        y_test_two_space_discrete = signal_utility.apply_discretization_two_states(y_test, splitting_threshold)

        #### two states discretization long only
        splitting_quantile_threshold_lo, splitting_threshold_lo, sharpe_strat_lo, sharpe_under_lo = signal_utility.find_best_two_states_discretization(
            y_train_pred, y_train)
        y_pred_two_space_discrete_lo = signal_utility.apply_discretization_two_states_lo(y_pred, splitting_threshold_lo)
        y_test_two_space_discrete_lo = signal_utility.apply_discretization_two_states_lo(y_test, splitting_threshold_lo)

        run_importances=forecasting_model.get_features_importance(features_names)
        features_importances.append(run_importances)
        # no leverage for the moment
        y_pred = np.clip(y_pred, a_min=min_signal, a_max=max_signal)
        forecasting_series.loc[slice_s] = y_pred.ravel()
        discrete_three_states_forecasting_series.loc[slice_s] = y_pred_three_space_discrete.ravel()
        discrete_two_states_forecasting_series.loc[slice_s] = y_pred_two_space_discrete.ravel()
        discrete_three_states_forecasting_series_lo.loc[slice_s] = y_pred_three_space_discrete_lo.ravel()
        discrete_two_states_forecasting_series_lo.loc[slice_s] = y_pred_two_space_discrete_lo.ravel()

        rmse = mean_squared_error(y_test, y_pred)
        matrix_three = confusion_matrix(y_test_three_space_discrete, y_pred_three_space_discrete)
        accuracy_three = accuracy_score(y_test_three_space_discrete, y_pred_three_space_discrete)
        # matrix_three_lo = confusion_matrix(y_test_three_space_discrete_lo, y_pred_three_space_discrete_lo)
        # accuracy_three_lo = accuracy_score(y_test_three_space_discrete_lo, y_pred_three_space_discrete_lo)
        matrix_two = confusion_matrix(y_test_two_space_discrete, y_pred_two_space_discrete)
        accuracy_two = accuracy_score(y_test_two_space_discrete, y_pred_two_space_discrete)
        matrix_two_lo = confusion_matrix(y_test_two_space_discrete_lo, y_pred_two_space_discrete_lo)
        accuracy_two_lo = accuracy_score(y_test_two_space_discrete_lo, y_pred_two_space_discrete_lo)

        if display:
            print('training slice ' + str(slice_n))
            print('testing slice ' + str(slice_s))
            print('rmse for slice : '+str(rmse))
            print('accuracy three states')
            print(accuracy_three)
            print('confusion matrix three states')
            print(matrix_three  )
            print('accuracy two states')
            print(accuracy_two)
            print('confusion matrix two states')
            print(matrix_two)
        iteration_counter = iteration_counter+1

    features_importances = pd.DataFrame(features_importances)
    return forecasting_series, features_importances, discrete_two_states_forecasting_series, discrete_two_states_forecasting_series_lo, discrete_three_states_forecasting_series, discrete_three_states_forecasting_series_lo

